
#include <cstdlib>
#include <cstdio>
#include <ctime>
#include <string>
#include <cstring>

using namespace std;


template <class T>
class ArrList {

    T* arr;
    int len;
    int head;
    int tail;

    // reallocate arr to ensure it can store at least n elements.
    void ensure_capacity(int n){
        // e.g. double the list capacity, if needed.
        if(len > n)     // check if we need to do this...
            return;             // no, there's enough space there.
        int newlen = len;
        while(newlen <= n)    // until we can store n elements.
            newlen = newlen * 2;    // keep on doubling list length
        // allocate a new array of size newlen.
        T* newarr = new T[newlen];
        for(int i=0;i<size();i++)
            newarr[i] = get(i);       // copy old list into new list.
        delete[] arr;       // free old list.
        arr = newarr;       // setup new list.
        head = size();      // head is the end of the list.
        tail = 0;           // new list starts at 0.
        len = newlen;       // setup new length.
    }


public:

    ArrList(){
        len = 8;
        arr = new T[len];
        head = tail = 0;
    }

    // copy constructor (make a copy of an existing list)
    // ArrList a = b;   // a does not "exist" before assignment.
    ArrList(ArrList& c){
        len = c.len;
        arr = new T[len];
        head = tail = 0;
        for(int i=0;i<c.size();i++)
            push(c.get(i));
    }

    ~ArrList(){
        // clean up whatever we allocated in constructor
        delete[] arr;
        len = head = tail = 0;  // zero things out.
        arr = (T*)0;
    }

    // assignment operator; 
    //  ArrList a;   // a exists now.
    //  a = b;       // assignment operator
    ArrList& operator=(ArrList& c){    
        delete[] arr;       // free list that's already there.
        len = c.len;
        arr = new T[len];
        head = tail = 0;
        for(int i=0;i<c.size();i++)
            push(c.get(i));
    }

    // check if list is empty (head is same as tail).
    bool isEmpty(){
        return head == tail;
    }

    // returns number of elements in the list.
    int size(){
        return (head - tail + len) % len;
    }

    // get i-th element starting from tail (index 0).
    T get(int i){
        if(i < 0)   // if negative, return starting from end.
            return arr[ (head + len + i) % len ];
        return arr[ (tail + i) % len ];
    }

    // set n at ith location.
    void set(int i,T n){
        if(i < 0){
            arr[ (head + len + i) % len ] = n;
        }else{
            arr[ (tail + i) % len ] = n;
        }
    }

    /*
    // HW?; returns a copy of this list, sorted ascending.
    ArrList sort(){
        // use any sort... 
        // temp variable is type... T.
        // comparisons <  >.
        // don't index arr[] directly; use set/get functions.
        // e.g. arr[0] may not equal get(0).
    }
    */

    // push (add) an integer onto the list (at head).
    void push(T n){
        ensure_capacity(size() + 1);    // ensure we have enough space.
        arr[head] = n;      // insert at head.
        head = (head + 1) % len;        // move head to next slot.
    }

    // reverse of push.
    T pop(){
        if(size() <= 0)     // this is a *really* bad error. 
            return arr[0];  // we want to exit program if this happens.
        head = (head + len - 1) % len;  // slot before head index.
        return arr[head];           // return element under head.
    }

    // remove element from tail (similar to pop, except from other end).
    T shift(){
        T n = arr[tail];
        tail = (tail + 1) % len;        // move head to next slot.
        return n;
    }

    // insert at the tail of the list (similar to push).
    void unshift(T n){
        ensure_capacity(size() + 1);    // ensure we have enough space.
        tail = (tail + len - 1) % len;        // move tail 1 back.
        arr[tail] = n;      // insert at tail.
    }

    // splice (index, length, replacewith).
    ArrList& splice(int n){
        // remove all elements after index n.
        while(size() > n)
            pop();
        return *this;
    }
    
    // at location n, remove l elements.
    ArrList& splice(int n,int l){
        ArrList tlst;
        // remove all elements after index n.
        while(size() > n)
            tlst.push(pop());   // save everything after n.
        while(l-- > 0)
            tlst.pop();     // remove l saved elements.
        // put back the remainer.
        while(tlst.size()>0)
            push(tlst.pop());
        return *this;
    }

    // at location n, remove l elements, replace with lst.
    ArrList& splice(int n,int l,ArrList& lst){
        //printf("before splice: len=%d, siz=%d, head=%d, tail=%d\n",len,size(),head,tail);
        ArrList tlst;
        // remove all elements after index n.
        while(size() > n)
            tlst.push(pop());   // save everything after n.
        while(l-- > 0)
            tlst.pop();     // remove l saved elements.
        for(int i=0;i<lst.size();i++)   // replacement list.
            push(lst.get(i));
        // put back the remainer.
        while(tlst.size()>0)
            push(tlst.pop());
            // printf("after splice: len=%d, siz=%d, head=%d, tail=%d\n",len,size(),head,tail);            
        return *this;
    }
};


class Circle {
public:
    float r;
};



int main(){

    srand(time(NULL));

    ArrList<int> a;

    // insert a random number of random integers into the list.
    int n = rand() % 100;
    for(int i=0;i<n;i++)
        a.push(rand()%100);

    printf("list is(%d): ",a.size());
    for(int i=0;i<a.size();i++)
        printf("%d ",a.get(i));
    printf("\n");

    printf("last element is: %d\n",a.get(-1));

    // create a list of characters A-Z.
    ArrList<char> charr;
    for(char c='A';c <='Z';c++)
        charr.push(c);


    ArrList<string> b;
    for(int i=0;i<100;i++){
        string s = (string)""+charr.get(rand()%charr.size())+
                  charr.get(rand()%charr.size())+
                  charr.get(rand()%charr.size());
        b.push(s);
    }

    printf("list is(%d): ",b.size());
    for(int i=0;i<b.size();i++)
        printf("%s ",b.get(i).c_str());
    printf("\n");

    printf("last element is: %s\n",b.get(-1).c_str());

    ArrList<Circle> carr;

    for(int i=0;i<10;i++){
        Circle circle;
        circle.r = rand();
        carr.push(circle);
    }

    return 0;
}



